/*********************************************************************NVMH2****
Path:  C:\DEV\devrel\NV_SDK_4\DX8\CommonSrc
File:  SimpleObject8.cpp

Copyright (C) 1999, 2000 NVIDIA Corporation
This file is provided without support, instruction, or implied warranty of any
kind.  NVIDIA makes no guarantee of its fitness for a particular purpose and is
not liable under any circumstances for any damages or loss whatsoever arising
from the use or inability to use this file or items derived from it.

Comments:


******************************************************************************/



#include <assert.h>
#include <windows.h>

#include <d3d8.h>
#include <d3dx8.h>
#include <math.h>
#include <vector>
#include <algorithm>

using namespace std;

#include "SimpleObject8.h"
#include "NV_Common.h"
#include "NV_Error.h"


#define UL_PI      3.1415926535897932384626433832795028841971693993751
#define UL_SQRT_2  1.4142135623730950488016887242096980785696718753769
#define UL_PI2     (UL_PI/2.0)


#define BOUNDCHK(in, cmp, msg, quit) {if(in>=cmp){DoError(msg,quit);}}

//////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SimpleObject8::SimpleObject8()
{
	m_pVertices = NULL;
	m_pIndices	= NULL;

	m_wNumVerts = 0;
	m_wNumInd	= 0;
	m_PrimType  = D3DPT_TRIANGLELIST;

	m_bIsValid = false;
}

SimpleObject8::~SimpleObject8()
{
	UnAllocate();
}

//////////////////////////////////////////////////////////////////////

void SimpleObject8::UnAllocate()
{
	if( m_pVertices )
		free( m_pVertices	);
	if( m_pIndices	)
		free( m_pIndices	);

	m_pVertices = NULL;
	m_pIndices = NULL;

	m_wNumVerts = 0;
	m_wNumInd	= 0;
	m_PrimType  = D3DPT_TRIANGLELIST;

	m_bIsValid = false;
}


void SimpleObject8::AllocateMoreIndices( WORD nind )
{
	// copies old indices into new array

	if( nind <= m_wNumInd )
	{
		FDebug("simpleobject8 more indices: ask for more, not less!\n");
		assert( false );
		return;
	}

	
	WORD * pInd = (WORD *) calloc( nind, sizeof( WORD ) * nind );
	NULLASSERT( pInd, "SimpleObject8:: Can't allocate more indices!", true );

	// copy old indices to new array
	memcpy( pInd, m_pIndices, sizeof(WORD) * m_wNumInd );

	// fill remaining indices with 0
	memset( & (pInd[ m_wNumInd ]), 0, sizeof(WORD) * (nind - m_wNumInd ) );

	m_wNumInd = nind;
	free( m_pIndices );

	m_pIndices = pInd;
}



void SimpleObject8::Allocate( WORD nvert, WORD nind )
{
	UnAllocate();

	m_pVertices = (SimpleObjVertex*) calloc( nvert, sizeof( SimpleObjVertex ) );
	NULLASSERT( m_pVertices, "SimpleObject8:: Can't allocate vertices!", true );
	m_wNumVerts = nvert;

	m_pIndices = (WORD*) calloc( nind, sizeof( WORD ) );
	NULLASSERT( m_pIndices, "SimpleObject8:: Can't allocate indices!", true );
	m_wNumInd = nind;

}


void SimpleObject8::FreeDuplicatePositions( D3DXVECTOR3 ** pPos )
{
	if( *pPos != NULL )
	{
		free( *pPos );
		*pPos = NULL;
	}	
}


void SimpleObject8::DuplicatePositions( D3DXVECTOR3 ** pPos, WORD * num_verts )
{
	// When done with the buffer, call the FreeDuplicatePositions() function
	//  above.

	*pPos = (D3DXVECTOR3*) calloc( m_wNumVerts, sizeof( D3DXVECTOR3 ));
	assert( *pPos != NULL );

	*num_verts = m_wNumVerts;

	int i;
	for( i=0; i < m_wNumVerts; i++ )
	{
		(*pPos)[i] = m_pVertices[i].pos;
	}
}


void SimpleObject8::FreeDuplicateIndices( WORD ** pInd )
{
	if( *pInd != NULL )
	{
		free( *pInd );
		*pInd = NULL;
	}
}

void SimpleObject8::DuplicateIndices( WORD ** pInd, WORD * old_n_ind )
{
	*pInd = (WORD*) calloc( m_wNumInd, sizeof( WORD ));
	assert( *pInd != NULL );

	*old_n_ind = m_wNumInd;

	memcpy( *pInd, m_pIndices, sizeof( WORD ) * m_wNumInd );
}



//-----------------------------------------------------------------------------
// Name: SimpleObject8::InitSphere
// Desc: Make a sphere with arbitrary basis axes.
//   *** There are no checks or restrictions forcing the basis axes to be
//       orthogonal!
//  The sphere is made by sweeping an arc, so it will have a seam of overlapping
//    unconnected vertices along one edge.  The cap is tesselated separately
//    to avoid zero-area triangles and to minimize distortion of the texture
//    at the poles.
//  The resulting sphere will have many vertices sharing the same position
//    at the poles.  These vertices each have different texture coordinates, 
//    and this makes for a sphere relatively free of severe distortions of 
//    texture at the poles that you would see with other approaches.
//
// ****** This will not generate proper normals if the 
//  axis vectors introduce a scale factor or are non-orthogonal
//  This generates correct normals only for a true sphere!
//-----------------------------------------------------------------------------

void SimpleObject8::InitSphere( float radius, int nLat, int nLong,
								const D3DXVECTOR3 & axis1, 
								const D3DXVECTOR3 & axis2, 
								const D3DXVECTOR3 & axis3			)
{
	int i,j;

	if( nLong < 1 )	nLong = 1;
	if( nLat  < 3 ) nLat  = 3;
	if( radius < 0 ) radius = -radius;

	assert( radius != 0.0f );


	//  If input nLong = 3, nLat = 2 (which are expanded for 
	//    vertex counts of nLong=4, nLat = 4, then the grid of
	//    vertices looks like this:
	//		
	//		  12   13    14   15
	//
	//		8    9    10   11
	//
	//		4    5    6    7
	//
	//		  0    1    2     3
	//
	//	** Vertices 15 and 3 are never used. **
	//  Vertices 0,1,2,3 are all at the same position at the pole.
	//  Same for vertices 12, 13, 14, 15
	//  The UV coords are offset by 1/2 the increment from 4 to 5
	//   to lessen distortion of textures at the poles.  If this
	//   were not done, the texture would spiral in at the poles
	//   in two rougly fan-like halves.
	//

	nLong++;		// repeat edge for texture coord happy
	nLat+=2;		// two rings on the ends

	WORD nvert = nLong * nLat;	

	////////////////////////////////////////////////
	// Explanation of ntri number:
	//
	// 2 * nLong * nLat => quads over grid nLong * nLat
	// - nLat * 2       => No quad for last row of verts which have
	//						same position as first row of verts
	// - nLong          => - 2 * nLong / 2 => Only 1 tri at end caps
	//					    instead of a quad

	WORD ntri  = 2 * nLong * nLat - nLong - nLat * 2;

	m_PrimType  = D3DPT_TRIANGLELIST;

	// Allocate memory:
	Allocate( nvert, ntri * 3 );

	//m_csDesc = "I'm a sphere!";
	//m_csName = "Sphere01";

	// make vertex positions
	int   index;
	float ringz;			// z cord of a latitude ring
	float ring_r;			// radius of x,y ring
	float theta, rho;		// theta = x,y plane angle, rho = z
	float x,y,u,v;				// x,y coord of pt.
	

	for( j=0; j < nLat; j++ )
	{
		rho = (float)( UL_PI * j / (float)(nLat-1) - UL_PI/2.0f );
		ringz  = (float) ( radius * sin(rho) );
		ring_r = (float) sqrt( fabs( radius * radius - ringz*ringz) );

		for( i=0; i < nLong; i++)
		{
			theta = (float)( 2* UL_PI * i / (float)( nLong - 1 ) );
			x = (float) ( ring_r * cos(theta) );
			y = (float) ( ring_r * sin(theta) );

			index = i + j*nLong;
			if( index > m_wNumVerts )
			{
				FDebug("InitSphere Error! %n index of only %n allocated!\n", index, m_wNumVerts );
				index = m_wNumVerts -1;
			}

			m_pVertices[ index ].pos.x = x * axis1.x + y * axis2.x + ringz * axis3.x;
			m_pVertices[ index ].pos.y = x * axis1.y + y * axis2.y + ringz * axis3.y;
			m_pVertices[ index ].pos.z = x * axis1.z + y * axis2.z + ringz * axis3.z;
				
			/*
			if( j==0 || j== 1 || j==nLat-1 )
			{
				FMsg("j=%4.4d     %f   %f   %f\n", j, m_pVertices[ index ].pos.x,
						m_pVertices[ index ].pos.y, m_pVertices[ index ].pos.z );

				FMsg("j=%4.4d     %f   %f   %f\n", j, x,y,ringz );
				FMsg("ringr   %f    theta %f\n", ring_r, theta );
				FMsg("radius  %f    \n", radius );
				FMsg("\n");
			}
			*/

			u = (float)( (float)i/(float)(nLong-1) );
			v = (float)( (float)j/(float)(nLat-1)  );

			// Correct for sphere pinch & "degenerate" triangle at caps
			if( j==0 )
			{
				// If at top/bottom of sphere
				u -= 0.5f / (float)nLong;
			}
			else if( j == nLat-1 )
			{
				// If at top/bottom of sphere
				u += 0.5f / (float)nLong;
			}

			m_pVertices[ index ].t0.x = u;
			m_pVertices[ index ].t0.y = v;

			//FDebug("U:  %f    V:  %f   ind: %d\n", u, v, index );

			//////////////////////////////////////////////////////////////////////
			// ****** This will not generate proper normals if the 
			//  axis vectors introduce a scale factor or are non-orthodonal
			//  This generates correct normals only for a true sphere!

			m_pVertices[index].nrm = m_pVertices[index].pos;

			D3DXVec3Normalize( & m_pVertices[index].nrm, &m_pVertices[index].nrm );
			//////////////////////////////////////////////////////////////////////


			m_pVertices[ index ].diffuse = D3DCOLOR_ARGB( 255, 255, 255, 255 );

		}
	}
	
	assert( index + 1 == m_wNumVerts );


	// stitch the tris
	// Culling is CCW
/*
//gj  This method builds zero-area tris and stitches the seam where
//    two vertices have the same position.  These tris are not needed.

	WORD  nind = 0;
	WORD  i0, i1, i2, i3;

	for( j=0; j < nLat-1; j++ )
	{
		for( i=0; i < nLong; i++)
		{
			i0 = i   + j*nLong;
			i1 = i   + (j+1)*nLong;
			i2 = (i+1)%nLong + j*nLong;		// % to wrap back to zero
			i3 = (i+1)%nLong + (j+1)*nLong;
		
			BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Index too high!", true );

			m_pIndices[nind++] = i2;
			m_pIndices[nind++] = i1;
			m_pIndices[nind++] = i0;

			BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Index too high 2!", true );
			m_pIndices[nind++] = i3;
			m_pIndices[nind++] = i1;
			m_pIndices[nind++] = i2;
		}
	}
*/
	/////////////////////////////////////////////////////////////////
	// This is a more efficient tesselation.  There are no zero-area
	//   tris

	// stitch the tris
	// Culling is CCW

	WORD  nind = 0;
	WORD  i0, i1, i2, i3;

	for( j=1; j < nLat-2; j++ )
	{
		for( i=0; i < nLong - 1; i++)
		{
			i0 = i   + j*nLong;
			i1 = i   + (j+1)*nLong;
			i2 = (i+1)%nLong + j*nLong;		// % to wrap back to zero
			i3 = (i+1)%nLong + (j+1)*nLong;
		
			BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Index too high!", true );

			m_pIndices[nind++] = i2;
			m_pIndices[nind++] = i1;
			m_pIndices[nind++] = i0;

			BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Index too high 2!", true );

			m_pIndices[nind++] = i3;
			m_pIndices[nind++] = i1;
			m_pIndices[nind++] = i2;
		}
	}

	// Stitch end caps
	// nLong-1 as there is no tri for last column of verts which is wrapped
	///  to lie at the same points as the first column.

	// Bottom cap:

	for( i=0; i < nLong-1; i++ )
	{
		// ^	nLong->
		// |

		// n
		// L
		// a
		// t
		//		  12   13    14   15
		//
		//		8    9    10   11
		//
		//		4    5    6    7
		//
		//		  0    1    2     3

		// First step = starts with vertex 0

		i0 = i          + 0 * nLong;

		// First step = vert 5
		// No need to modulo to wrap back - only 
		// stitching nLong-1 tris (3 for the above case)
		i1 = (i+1)		+ 1 * nLong;

		// First step:  end with vert 4

		i2 = i			+ 1 * nLong;


		BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Capping Index too high!", true );
		m_pIndices[ nind++ ] = i0;
		m_pIndices[ nind++ ] = i1;
		m_pIndices[ nind++ ] = i2;
	}

	// Now for the top cap
	for( i=0; i < nLong-1; i++ )
	{
		//		  12   13    14   15
		//
		//		8    9    10   11
		//
		//		4    5    6    7
		//
		//		  0    1    2     3

		// First step = starts with vertex 12

		i0 = i          + (nLat-1) * nLong;

		// First step = vert 8

		i1 = i			+ (nLat-2) * nLong;

		// First step:  last vert = 9

		i2 = (i+1)		+ (nLat-2) * nLong;

		BOUNDCHK( nind, m_wNumInd-2, "InitSphere:: Capping Index too high!", true );
		m_pIndices[ nind++ ] = i0;
		m_pIndices[ nind++ ] = i1;
		m_pIndices[ nind++ ] = i2;
	}

	m_wNumInd = nind;

	/////////////////////////////////////////////////////////////////

	m_bIsValid = true;
	PostInit();

	// all done
}						// InitSphere




//-----------------------------------------------------------------------------
// Name: SimpleObject8::InitSphere( float radius, int nLat, int nLong )
// Desc: Make a sphere using a default set of axes.  
//-----------------------------------------------------------------------------

void SimpleObject8::InitSphere( float radius, int nLat, int nLong )
{
	InitSphere( radius, nLat, nLong,
				D3DXVECTOR3( 1.0f, 0.0f, 0.0f ),
				D3DXVECTOR3( 0.0f, 1.0f, 0.0f ),
				D3DXVECTOR3( 0.0f, 0.0f, 1.0f )  );
}

///////////////////////////////////////////////////////////////////////////////


void SimpleObject8::InitTest1_4QuadPlane(	const D3DXVECTOR3 & center,
											const D3DXVECTOR3 & axis1,
											const D3DXVECTOR3 & axis2  )
{

	Allocate( 16, 24 );
	m_PrimType = D3DPT_TRIANGLELIST;
	m_bIsValid = true;

	int i;
	i=0;

	D3DXVECTOR3  d1 = axis1 / 2.0f;
	D3DXVECTOR3  d2 = axis2 / 2.0f;

	m_pVertices[0].pos = center;
	m_pVertices[1].pos = center + d1;
	m_pVertices[2].pos = center + d1 + d2;
	m_pVertices[3].pos = center + d2;

	m_pIndices[i++] = 0;	m_pIndices[i++] = 1;	m_pIndices[i++] = 2;
	m_pIndices[i++] = 0;	m_pIndices[i++] = 2;	m_pIndices[i++] = 3;

	m_pVertices[4].pos = center;
	m_pVertices[5].pos = center + d2;
	m_pVertices[6].pos = center + d2 - d1;
	m_pVertices[7].pos = center - d1;

	m_pVertices[4 ].diffuse = 0x0000FF00;
	m_pVertices[5 ].diffuse = 0x0000FF00;
	m_pVertices[6 ].diffuse = 0x0000FF00;
	m_pVertices[7 ].diffuse = 0x0000FF00;

	int inc;
	inc = 4;
	m_pIndices[i++] = m_pIndices[0] + inc;	m_pIndices[i++] = m_pIndices[1] + inc;
	m_pIndices[i++] = m_pIndices[2] + inc;	m_pIndices[i++] = m_pIndices[3] + inc;
	m_pIndices[i++] = m_pIndices[4] + inc;	m_pIndices[i++] = m_pIndices[5] + inc;
	
	m_pVertices[8 ].pos = center;
	m_pVertices[9 ].pos = center - d1;
	m_pVertices[10].pos = center - d2 - d1;
	m_pVertices[11].pos = center - d2;

	m_pVertices[8 ].diffuse = 0x00FF0000;
	m_pVertices[9 ].diffuse = 0x00FF0000;
	m_pVertices[10].diffuse = 0x00FF0000;
	m_pVertices[11].diffuse = 0x00FF0000;

	inc = 8;
	m_pIndices[i++] = m_pIndices[0] + inc;	m_pIndices[i++] = m_pIndices[1] + inc;
	m_pIndices[i++] = m_pIndices[2] + inc;	m_pIndices[i++] = m_pIndices[3] + inc;
	m_pIndices[i++] = m_pIndices[4] + inc;	m_pIndices[i++] = m_pIndices[5] + inc;

	m_pVertices[12].pos = center;
	m_pVertices[13].pos = center - d2;
	m_pVertices[14].pos = center - d2 + d1;
	m_pVertices[15].pos = center + d1;

	m_pVertices[12].diffuse = 0x000000FF;
	m_pVertices[13].diffuse = 0x000000FF;
	m_pVertices[14].diffuse = 0x000000FF;
	m_pVertices[15].diffuse = 0x000000FF;

	inc = 12;
	m_pIndices[i++] = m_pIndices[0] + inc;	m_pIndices[i++] = m_pIndices[1] + inc;
	m_pIndices[i++] = m_pIndices[2] + inc;	m_pIndices[i++] = m_pIndices[3] + inc;
	m_pIndices[i++] = m_pIndices[4] + inc;	m_pIndices[i++] = m_pIndices[5] + inc;

	assert( i == 24 );

	GenerateNormalsCCW();
	PostInit();
}


// ============================================================ //
// Function name	: SimpleObject8::InitRectangularBlock_Smooth
// Description	    : 
//		This function really makes a parallel piped with 8 vertices
//      Corners have only one normal so the object is "smooth"
//      Use the _Faceted(..) to create a block with degenerate 
//  vertices for different normals at the corners
//		axis1, 2, and 3 should specify a right 
//  handed coordinate system, where 1 CROSS 2 points in the
//  direction of 3 for a right-handed cross product.  The function
//	does not check for this.
//      Length of each side will be the lenght of each axis vector
//  Winding is CCW
//  Color is ARGB
//
// ============================================================ //

void SimpleObject8::InitRectangularBlock_Smooth( const D3DXVECTOR3 center,
												 DWORD color,
												 const D3DXVECTOR3 & axis1,
												 const D3DXVECTOR3 & axis2,
												 const D3DXVECTOR3 & axis3  )
{
	int i;
	int nvert, nind;

	nvert = 8;
	nind  = 12 * 3;

	Allocate( nvert, nind );

	m_PrimType = D3DPT_TRIANGLELIST;

	D3DXVECTOR3		pVert[ 8 ];

	pVert[0] = center - axis1 / 2.0f - axis2 / 2.0f - axis3 / 2.0f;
	pVert[1] = pVert[0] + axis1;
	pVert[2] = pVert[1] + axis2;
	pVert[3] = pVert[0] + axis2;

	for( i=4; i < 8; i++ )
	{
		pVert[i] = pVert[ i - 4 ] + axis3;
	}

	for( i=0; i < 8; i++ )
	{
		m_pVertices[i].pos = pVert[i];
		m_pVertices[i].t0 = D3DXVECTOR2( 0.0f, 0.0f );
		m_pVertices[i].diffuse = color;
	}

	i=0;
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 1;	
	m_pIndices[i++] = 5;	
							
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 5;	
	m_pIndices[i++] = 4;	
							
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 4;	
	m_pIndices[i++] = 7;	
							
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 7;	
	m_pIndices[i++] = 3;	
							
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 2;	
	m_pIndices[i++] = 1;	
							
	m_pIndices[i++] = 0;	
	m_pIndices[i++] = 3;	
	m_pIndices[i++] = 2;	

	//////////////////////////////

	m_pIndices[i++] = 7;	
	m_pIndices[i++] = 4;	
	m_pIndices[i++] = 5;	
							
	m_pIndices[i++] = 7;	
	m_pIndices[i++] = 5;	
	m_pIndices[i++] = 6;	
							
	m_pIndices[i++] = 6;	
	m_pIndices[i++] = 5;	
	m_pIndices[i++] = 2;	
							
	m_pIndices[i++] = 2;	
	m_pIndices[i++] = 5;	
	m_pIndices[i++] = 1;	
							
	m_pIndices[i++] = 2;	
	m_pIndices[i++] = 7;	
	m_pIndices[i++] = 6;	
							
	m_pIndices[i++] = 2;	
	m_pIndices[i++] = 3;	
	m_pIndices[i++] = 7;	

	m_bIsValid = true;

	GenerateNormalsCCW();
	PostInit();
}



// ============================================================= //
// Function name	: SimpleObject8::InitRectangularBlock_Faceted
// Description	    : Similar to the _Smooth but this one makes
//    a block with 24 vertices instead of only 8.  This is so 
//    each vertex can have a different normal and the cube can
//    be rendered with facets instead of smooth shading.
//
// ============================================================= //

void SimpleObject8::InitRectangularBlock_Faceted( const D3DXVECTOR3 center,
												  DWORD color,
												  const D3DXVECTOR3 & axis1,
												  const D3DXVECTOR3 & axis2,
												  const D3DXVECTOR3 & axis3  )
{

	InitRectangularBlock_Smooth( center, color, axis1, axis2, axis3 );

	// Take the vertices / tris from the smooth block and expand them out into
	//  independent faces

	MakeFaceted();				// create degenerate vertices for each triangle

	GenerateNormals();

	m_bIsValid = true;

	PostInit();
}


// ============================================================= //
// Function name	: SimpleObject8::MakeFaceted
// Description	    : Break up every triangle to use its own 
//    unique vertices, so every face will have a unique normal --
//    a flat face normal across the entire face.
//   This will increase the vertex count, to say the least.
// ============================================================= //


void SimpleObject8::MakeFaceted()
{
	int i;

	WORD nvert = m_wNumInd;


	SimpleObject8 * pClone = new SimpleObject8;
	assert( pClone != NULL );

	pClone->InitClone( this );


	// Allocate new vertices
	Allocate( nvert, m_wNumInd );


	// Fill in the unique vertices for each triangle from the
	//  cloned object

	for( i = 0; i < m_wNumVerts; i++ )
	{
		m_pVertices[i] = pClone->m_pVertices[ pClone->m_pIndices[i] ];
	}

	// re-map indices to use the new vertices
	for( i = 0; i < m_wNumInd; i++ )
	{
		m_pIndices[i] = i;
	}

	m_bIsValid = true;

	// Do not regenerate normals here
	// User can do that after MakeFaceted()

	// Free the cloned object
	delete pClone;
	pClone = NULL;

}



void	SimpleObject8::ListVertIndicesUsed()
{
	// Text output to debug console

	V_WORD verts;
	V_WORD :: iterator v_end;

	int i;

	for(i=0; i < m_wNumInd; i++ )
	{
		verts.push_back( m_pIndices[ i ] );
	}

	sort( verts.begin(), verts.end() );
	v_end = unique( verts.begin(), verts.end() );

	verts.erase( v_end, verts.end() );

	FDebug("Vertex indices used for triangles:\n");
	for( i=0; i < verts.size(); i++ )
	{
		FDebug("%5.5d   ", verts.at(i) );

		if( (i+1)%10 == 0 )
			FDebug("\n");
	}

	FDebug("\n");
	FDebug("\n");
}


// ============================================================= //
// Function name	: SimpleObject8 TestEliminateTrisWithoutTheseVertices
// Description	    : Eliminate all faces which do not use one of the vertex
//    indices in the verts array.
//    V_WORD = { 0 }  eliminates all faces which do not use vertex 0
// ============================================================= //

void SimpleObject8::TestEliminateTrisWithoutTheseVertices( V_WORD & verts )
{
#define FUNC_DEBUG	0

	WORD i, nv;
	WORD verti;
	WORD tribase;

	V_WORD	tris_to_keep;
	tris_to_keep.reserve( verts.size() * 6 );

	// determine which tris to keep

	for( nv = 0; nv < verts.size(); nv++ )
	{
		verti = verts.at(nv);
		
		for( i=0; i < m_wNumInd; i += 3 )
		{
			if( m_pIndices[i]	== verti || 
				m_pIndices[i+1] == verti ||
				m_pIndices[i+2] == verti	)
			{
				//	FDebug("  Vertex %d found in tri %d\n", nv, i/3);
				tris_to_keep.push_back( i/3 );
			}
		}
	}

#if FUNC_DEBUG
	FDebug("\n");
	FDebug("Tris to keep before unique:  size:  %d\n", tris_to_keep.size() );
	for( i=0; i < tris_to_keep.size(); i++ )
	{
		FDebug("%4.4d   ", tris_to_keep.at(i));
		if( (i+1) % 10 == 0 )
			FDebug("\n");		
	}
	FDebug("\n\n");
#endif

	V_WORD::iterator t_end;

	sort( tris_to_keep.begin(), tris_to_keep.end() );

	// unique doesn't reduce size of the array, so need to 
	//  track the end thing it returns.
	t_end = unique( tris_to_keep.begin(), tris_to_keep.end() );

	tris_to_keep.erase( t_end, tris_to_keep.end() );

#if FUNC_DEBUG
	FDebug("\n");
	FDebug("Tris to keep after unique:  size:  %d\n", tris_to_keep.size() );
	for( i=0; i < tris_to_keep.size(); i++ )
	{
		FDebug("%4.4d   ",  tris_to_keep.at(i) );
		if( (i+1) % 10 == 0 )
			FDebug("\n");		
	}
	FDebug("\n\n");
#endif

	nv = 0;

	// Copy tris into smaller set of indices - do not reallocate
	//  the index array, change only the "last index" m_wNumInd value

	for( i=0; i < tris_to_keep.size(); i++ )
	{
		tribase = ( tris_to_keep.at(i) ) * 3;
		assert( tribase + 2 < m_wNumInd );
		assert( nv+2        < m_wNumInd );

		m_pIndices[nv++] = m_pIndices[ tribase     ];
		m_pIndices[nv++] = m_pIndices[ tribase + 1 ];
		m_pIndices[nv++] = m_pIndices[ tribase + 2 ];
	}

	m_wNumInd = nv;

}



void SimpleObject8::InitSpiral( const D3DXVECTOR3 & width_direction, 
								const D3DXVECTOR3 & height_direction,
								float width, float height, int n_sub_width, 
								int n_sub_height, float n_twists )
{

	int i,j;
	int nvert, nind;

	if( n_sub_height < 0 )
		n_sub_height = 0;
	if( n_sub_width < 0 )
		n_sub_width = 0;

	D3DXVECTOR3	width_dir;
	D3DXVec3Normalize( &width_dir, &width_direction );
	D3DXVECTOR3	height_dir;
	D3DXVec3Normalize( &height_dir, &height_direction );
	D3DXVECTOR3 length_dir;
	D3DXVec3Cross( &length_dir, &width_dir, &height_dir );



	nvert = (2 + n_sub_width)*(n_sub_height + 2);
	nind = (n_sub_width+1)*(n_sub_height+1)*2 * 3;

	Allocate( nvert, nind );

	m_PrimType = D3DPT_TRIANGLELIST;

	float z;
	float dz, dx;
	float angle;
	float da;
	float radius;

	z = -height / 2.0f;
	angle = 0;
	
	dz = height / (n_sub_height+1.0f);
	da = (float)( n_twists * UL_PI * 2.0f ) / (n_sub_height + 1.0f);

	dx =  width / ( (float)n_sub_width + 1.0f );


	int ind;

	float wt, ht, lt;		// params for each position axis

	for( j=0; j < n_sub_height + 2; j++ )
	{
		for(i=0; i < n_sub_width + 2; i++)
		{
			radius = width/2.0f - dx * i;

			ind = i + j * (n_sub_width+2);

			assert( ind < nvert );

			wt = radius * (float) sin( angle ); // amount of position in width_dir direction
			lt = radius * (float) cos( angle );	// amount of position in height_dir direction
			ht = z;

			m_pVertices[ ind ].pos.x = wt * width_dir.x + ht * height_dir.x + lt * length_dir.x;
			m_pVertices[ ind ].pos.y = wt * width_dir.y + ht * height_dir.y + lt * length_dir.y;
			m_pVertices[ ind ].pos.z = wt * width_dir.z + ht * height_dir.z + lt * length_dir.z;

			m_pVertices[ ind ].t0.x = (float)i/(n_sub_width  +1.0f);
			m_pVertices[ ind ].t0.y = (float)j/(n_sub_height +1.0f);

		}

		z += dz;
		angle += da;
	}

	// Set the indices
	int triind;				// triind = index of triangle -> 3 for each
							//   so 0,1,2 = 1st tri, 3,4,5 = 2nd tri
	triind = 0;

	for( j=0; j < n_sub_height + 1; j++ )
	{
		for( i=0; i < n_sub_width + 1; i++ )
		{

			if( i + j*(n_sub_width + 1) >= nvert )
				assert( false );
			if( i + 1 + j * ( n_sub_width + 2 ) >= nvert )
				assert( false );

			if( triind >= nind )
			{
				FDebug( "bad index calc: %d of %d", triind, nind );
				assert( false );
			}

			
			// stitching pattern is:
			//						           u,v = 1,1
			//			  v2 -- v1	      v2
			//				| /		    /  |
			//				|/		   /   |
			//            v0		v0 -- v1
			// u,v = 0,0

			assert( triind < nind );
			m_pIndices[ triind++ ] = i + j*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + (j+1)*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + j*(n_sub_width+2);
			assert( triind < nind );

			m_pIndices[ triind++ ] = i + j*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + (j+1)*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + (j+1)*(n_sub_width+2);		
		}
	}

	m_bIsValid = true;
	PostInit();
}


//@@@@@ fix to call other initspiral

void SimpleObject8::InitSpiral( float width, float height, int n_sub_width,
					int n_sub_height, float n_twists )
{

	int i,j;
	int nvert, nind;

	if( n_sub_height < 0 )
		n_sub_height = 0;
	if( n_sub_width < 0 )
		n_sub_width = 0;


	nvert = (2 + n_sub_width)*(n_sub_height + 2);
	nind = (n_sub_width+1)*(n_sub_height+1)*2 * 3;

	Allocate( nvert, nind );

	m_PrimType = D3DPT_TRIANGLELIST;

	float z;
	float dz, dx;
	float angle;
	float da;
	float radius;

	z = -height / 2.0f;
	angle = 0;
	
	dz = height / (n_sub_height+1.0f);
	da = (float)( n_twists * UL_PI * 2.0f ) / (n_sub_height + 1.0f);

	dx =  width / ( (float)n_sub_width + 1.0f );


	int ind;

	for( j=0; j < n_sub_height + 2; j++ )
	{
		for(i=0; i < n_sub_width + 2; i++)
		{
			radius = width/2.0f - dx * i;

			ind = i + j * (n_sub_width+2);

			assert( ind < nvert );

			m_pVertices[ ind ].pos.x = radius * (float) sin( angle );
			m_pVertices[ ind ].pos.y = radius * (float) cos( angle );
			m_pVertices[ ind ].pos.z = z;

			m_pVertices[ ind ].t0.x = (float)i/(n_sub_width  +1.0f);
			m_pVertices[ ind ].t0.y = (float)j/(n_sub_height +1.0f);

		}

		z += dz;
		angle += da;
	}

	// Set the indices
	int triind;				// triind = index of triangle -> 3 for each
							//   so 0,1,2 = 1st tri, 3,4,5 = 2nd tri
	triind = 0;

	for( j=0; j < n_sub_height + 1; j++ )
	{
		for( i=0; i < n_sub_width + 1; i++ )
		{

			if( i + j*(n_sub_width + 1) >= nvert )
				assert( false );
			if( i + 1 + j * ( n_sub_width + 2 ) >= nvert )
				assert( false );

			if( triind >= nind )
			{
				FDebug( "bad index calc: %d of %d", triind, nind );
				assert( false );
			}

			
			// stitching pattern is:
			//						x,z = 1,1
			//			  v2 -- v1	      v2
			//				| /		    /  |
			//				|/		   /   |
			//            v0		v0 -- v1
			// x,z = 0,0

			assert( triind < nind );
			m_pIndices[ triind++ ] = i + j*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + (j+1)*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + j*(n_sub_width+2);
			assert( triind < nind );

			m_pIndices[ triind++ ] = i + j*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + (j+1)*(n_sub_width+2);
			assert( triind < nind );
			m_pIndices[ triind++ ] = i + 1 + (j+1)*(n_sub_width+2);		
		}
	}

	m_bIsValid = true;
	PostInit();
}



void SimpleObject8::InitTorusWrapped( float InnerRadius, float OuterRadius, UINT Sides, UINT Rings )
{
	// Creates a torus "spun" about the z axis.
	// This one has no overlapping vertices on the seams,
	//  so you must use texture coordinate wrapping when 
	//  drawing it.

	UINT i,j;
	int nvert, nind;

	if( Sides < 3 )
		Sides = 3;
	if( Rings < 3 )
		Rings = 3;
	if( InnerRadius > OuterRadius )
	{
		float tmp = InnerRadius;
		InnerRadius = OuterRadius;
		OuterRadius = InnerRadius;
	}

	nvert = Sides * Rings;
	nind = 3 * Sides * 2 * Rings;		// # of triangle vertex indices

	// allocate memory for vertices
	Allocate( nvert, nind );


	m_PrimType = D3DPT_TRIANGLELIST;	// Could optimize to create tri
										//  strips, but not today!


	// Fill in vertex data

	// phi is angle around z axis
	// theta is angle around each ring

	float d_phi = 2.0f * (float)UL_PI / ((float)Rings);
	float d_theta = 2.0f * (float)UL_PI / ((float)Sides);

	float cent_radius = (InnerRadius + OuterRadius) / 2.0f;
	float ring_radius = (OuterRadius - InnerRadius) / 2.0f;

	float angle;


	SimpleObjVertex * pRing = (SimpleObjVertex*) calloc( Sides, sizeof( SimpleObjVertex));
	NULLASSERT( pRing, "SimpleObject8:: Can't allocate a ring of vertices!", true );


	// make the first ring - centered about cent_radius;
	for( i=0; i < Sides; i++ )
	{
		angle = 2.0f * (float)UL_PI * (float)i / (float)Sides;

		pRing[i].pos.x = ring_radius * ((float)cos( angle ));
		pRing[i].pos.z = ring_radius * ((float)sin( angle ));
		pRing[i].pos.y = 0.0f;

		pRing[i].pos.x += cent_radius;

		// Set normals
		pRing[i].nrm.x = (float) cos( angle );
		pRing[i].nrm.z = (float) sin( angle );
		pRing[i].nrm.y = 0.0f;

		pRing[i].t0.x = (float)i / (float)Sides;
	}

	int vind;
	float nx, ny, nz, length;
	vind = 0;

	for( j=0; j < Rings; j++ )
	{
		// rotate each ring into position
		//  around the sweep angle
		
		angle = 2.0f * (float)UL_PI * (float)j / (float) Rings;

		for( i=0; i < Sides; i++, vind++ )
		{
			assert( vind < nvert );
			m_pVertices[ vind ].pos.x = pRing[i].pos.x * ((float)cos( angle ));
			m_pVertices[ vind ].pos.y = pRing[i].pos.x * ((float)sin( angle ));
			m_pVertices[ vind ].pos.z = pRing[i].pos.z;

			// rotate normals into position
			nx = m_pVertices[ vind ].nrm.x = pRing[i].nrm.x * ((float)cos(angle));
			ny = m_pVertices[ vind ].nrm.y = pRing[i].nrm.x * ((float)sin(angle));
			nz = m_pVertices[ vind ].nrm.z = pRing[i].nrm.z;

			// normalize normals.  Shouldn't have to in this case, but
			//  just a safety measure.
			length = (float) sqrt(nx*nx + ny*ny + nz*nz);
			assert( length > 0.0f );

			m_pVertices[ vind ].nrm.x = nx / length;
			m_pVertices[ vind ].nrm.y = ny / length;
			m_pVertices[ vind ].nrm.z = nz / length;

			// Set texture coords:
			m_pVertices[ vind ].t0.x = pRing[i].t0.x;
			m_pVertices[ vind ].t0.y = (float) j / (float) Rings;
		}
	}

	///////////////////////////////////////////
	// Now make indices to form triangles

	int ind;			// index location to fill with a vertex index
	int v1, v2, v3;		// vertex indices of a triangle
	UINT jplus, iplus;	// j+1, i+1 wrapped back to 0 if > Rings or Sides, respectively


	ind = 0;

	for( j=0; j < Rings; j ++ )
	{
		for( i=0; i < Sides; i++ )
		{
			// first triangle of each "square" patch of surface
			//  on the torus

			iplus = i + 1;
			if( iplus == Sides )
				iplus = 0;
			jplus = j + 1;
			if( jplus == Rings )
				jplus = 0;


			// CCW winding order
			v1 = j		* Sides		+ i;
			v2 = j		* Sides		+ iplus;
			v3 = jplus	* Sides		+ i;


			assert( v1 < nvert && v1 >= 0 );
			assert( v2 < nvert && v2 >= 0 );
			assert( v3 < nvert && v3 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;

			// second triangle of the pair
			v1 = v3;
			v2 = v2;
			v3 = jplus * Sides	+ iplus;

			assert( v3 < nvert && v1 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;

		}
	}



	//////////////////////////////////////////

	if( pRing != NULL )
	{
		free( pRing );
		pRing = NULL;
	}

	m_bIsValid = true;
	PostInit();

}


/*
void SimpleObject8::InitTorus( float InnerRadius, float OuterRadius, UINT Sides, UINT Rings,
								float TexCoordUMax, float TexCoordVMax )
{

	// Create a torus with user-adjustable scaling of texture coords so a 
	//  texture can repeat several times around the torus

	InitTorus( InnerRadius, OuterRadius, Sides, Rings );

	int i;

	for( i=0; i < m_wNumVerts; i ++ )
	{
		m_pVertices[i].t0.x *= TexCoordUMax;
		m_pVertices[i].t0.y *= TexCoordVMax;
	}

	m_bIsValid = true;
	PostInit();

}

  */

void SimpleObject8::InitTorus( float InnerRadius, float OuterRadius, UINT Sides, UINT Rings,
								const D3DXVECTOR3 & axis1,
								const D3DXVECTOR3 & axis2,
								const D3DXVECTOR3 & axis3	)
{
	// Creates a torus "spun" about the 3rd axis.
	// With this non-wrapped model, vertices overlap at the seams.  
	// These overlapped vertices have appropriate texture coords so no
	//  change in tex coord wrapping mode is necessary.

	// There is no constraint saying tha the axes must be orthonormal
	// or consist of normalized vectors. 

	// *** This function will create incorrect normals if the three
	// axis vectors are not orthonormal or consist of non-normalized vectors


	UINT i,j;
	int nvert, nind;

	if( Sides < 3 )
		Sides = 3;
	if( Rings < 3 )
		Rings = 3;
	if( InnerRadius > OuterRadius )
	{
		float tmp = InnerRadius;
		InnerRadius = OuterRadius;
		OuterRadius = InnerRadius;
	}

	nvert = ( Sides + 1 ) * ( Rings + 1 );
	nind = 3 * Sides * 2 * Rings;		// # of triangle vertex indices

	// allocate memory for vertices
	Allocate( nvert, nind );


	m_PrimType = D3DPT_TRIANGLELIST;	// Could optimize to create tri
										//  strips, but not today!


	// Fill in vertex data

	// phi is angle around z axis
	// theta is angle around each ring

	UINT  nphi_verts;
	UINT  ntheta_verts;

	nphi_verts = Rings + 1;
	ntheta_verts = Sides + 1;

	float d_phi = 2.0f * (float)UL_PI / ((float)(nphi_verts - 1) );
	float d_theta = 2.0f * (float)UL_PI / ((float)(ntheta_verts - 1));

	float cent_radius = (InnerRadius + OuterRadius) / 2.0f;
	float ring_radius = (OuterRadius - InnerRadius) / 2.0f;

	float angle;


	SimpleObjVertex * pRing = (SimpleObjVertex*) calloc( ntheta_verts, sizeof( SimpleObjVertex));
	NULLASSERT( pRing, "SimpleObject8:: Can't allocate a ring of vertices!", true );


	// make the first ring - centered about cent_radius;
	for( i=0; i < ntheta_verts; i++ )
	{
		angle = (float) i * d_theta;

		pRing[i].pos.x = ring_radius * ((float)cos( angle ));
		pRing[i].pos.z = ring_radius * ((float)sin( angle ));
		pRing[i].pos.y = 0.0f;

		pRing[i].pos.x += cent_radius;

		// Set normals
		pRing[i].nrm.x = (float) cos( angle );
		pRing[i].nrm.z = (float) sin( angle );
		pRing[i].nrm.y = 0.0f;

		pRing[i].t0.x = (float) i / (float)(ntheta_verts - 1);
	}

	int vind;
	float nx, ny, nz, length;
	vind = 0;

	for( j=0; j < nphi_verts; j++ )
	{
		// rotate each ring into position
		//  around the sweep angle
		
		angle = (float) j * d_phi;

		for( i=0; i < ntheta_verts; i++, vind++ )
		{
			assert( vind < nvert );
			m_pVertices[ vind ].pos.x = pRing[i].pos.x * ((float)cos( angle ));
			m_pVertices[ vind ].pos.y = pRing[i].pos.x * ((float)sin( angle ));
			m_pVertices[ vind ].pos.z = pRing[i].pos.z;

			// rotate normals into position
			nx = m_pVertices[ vind ].nrm.x = pRing[i].nrm.x * ((float)cos(angle));
			ny = m_pVertices[ vind ].nrm.y = pRing[i].nrm.x * ((float)sin(angle));
			nz = m_pVertices[ vind ].nrm.z = pRing[i].nrm.z;

			// normalize normals.  Shouldn't have to in this case, but
			//  just a safety measure.
			length = (float) sqrt(nx*nx + ny*ny + nz*nz);
			assert( length > 0.0f );

			m_pVertices[ vind ].nrm.x = nx / length;
			m_pVertices[ vind ].nrm.y = ny / length;
			m_pVertices[ vind ].nrm.z = nz / length;

			// Set texture coords:
			m_pVertices[ vind ].t0.x = pRing[i].t0.x;
			m_pVertices[ vind ].t0.y = (float) j / (float)( nphi_verts - 1);
		}
	}

	///////////////////////////////////////////
	// Now make indices to form triangles

	int ind;			// index location to fill with a vertex index
	int v1, v2, v3;		// vertex indices of a triangle
	UINT jplus, iplus;	// j+1, i+1 wrapped back to 0 if > Rings or Sides, respectively


	ind = 0;

	for( j=0; j < Rings; j ++ )
	{
		for( i=0; i < Sides; i++ )
		{
			// first triangle of each "square" patch of surface
			//  on the torus

			iplus = i + 1;
			jplus = j + 1;

			// CCW winding order
			v1 = j		* ntheta_verts		+ i;
			v2 = j		* ntheta_verts		+ iplus;
			v3 = jplus	* ntheta_verts		+ i;


			assert( v1 < nvert && v1 >= 0 );
			assert( v2 < nvert && v2 >= 0 );
			assert( v3 < nvert && v3 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;

			// second triangle of the pair
			v1 = v3;
			v2 = v2;
			v3 = jplus * ntheta_verts	+ iplus;

			assert( v3 < nvert && v1 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;
		}
	}


	D3DXVECTOR3 param;
	D3DXVECTOR3 d1, d2, d3;

	d1.x = axis1.x; d1.y = axis2.x; d1.z = axis3.x;
	d2.x = axis1.y; d2.y = axis2.y; d2.z = axis3.y;
	d3.x = axis1.z; d3.y = axis2.z; d3.z = axis3.z;


	for( i=0; i < m_wNumVerts; i++ )
	{
		param = m_pVertices[ i ].pos;

		// rotate positions into new axis
		m_pVertices[ i ].pos.x = D3DXVec3Dot( & param, & d1 );
		m_pVertices[ i ].pos.y = D3DXVec3Dot( & param, & d2 );
		m_pVertices[ i ].pos.z = D3DXVec3Dot( & param, & d3 );

		// rotate normals into position
		param = m_pVertices[ i ].nrm;

		m_pVertices[ i ].nrm.x = D3DXVec3Dot( & param, & d1 );
		m_pVertices[ i ].nrm.y = D3DXVec3Dot( & param, & d2 );
		m_pVertices[ i ].nrm.z = D3DXVec3Dot( & param, & d3 );
	}


	//////////////////////////////////////////

	if( pRing != NULL )
	{
		free( pRing );
		pRing = NULL;
	}

	m_bIsValid = true;
	PostInit();

}



void SimpleObject8::InitTorus( float InnerRadius, float OuterRadius, UINT Sides, UINT Rings )
{
	// Creates a torus "spun" about the z axis.
	// With this non-wrapped model, vertices overlap at the seams.  
	// These overlapped vertices have appropriate texture coords so no
	//  change in tex coord wrapping mode is necessary.


	UINT i,j;
	int nvert, nind;

	if( Sides < 3 )
		Sides = 3;
	if( Rings < 3 )
		Rings = 3;
	if( InnerRadius > OuterRadius )
	{
		float tmp = InnerRadius;
		InnerRadius = OuterRadius;
		OuterRadius = InnerRadius;
	}

	nvert = ( Sides + 1 ) * ( Rings + 1 );
	nind = 3 * Sides * 2 * Rings;		// # of triangle vertex indices

	// allocate memory for vertices
	Allocate( nvert, nind );


	m_PrimType = D3DPT_TRIANGLELIST;	// Could optimize to create tri
										//  strips, but not today!


	// Fill in vertex data

	// phi is angle around z axis
	// theta is angle around each ring

	UINT  nphi_verts;
	UINT  ntheta_verts;

	nphi_verts = Rings + 1;
	ntheta_verts = Sides + 1;

	float d_phi = 2.0f * (float)UL_PI / ((float)(nphi_verts - 1) );
	float d_theta = 2.0f * (float)UL_PI / ((float)(ntheta_verts - 1));

	float cent_radius = (InnerRadius + OuterRadius) / 2.0f;
	float ring_radius = (OuterRadius - InnerRadius) / 2.0f;

	float angle;


	SimpleObjVertex * pRing = (SimpleObjVertex*) calloc( ntheta_verts, sizeof( SimpleObjVertex));
	NULLASSERT( pRing, "SimpleObject8:: Can't allocate a ring of vertices!", true );


	// make the first ring - centered about cent_radius;
	for( i=0; i < ntheta_verts; i++ )
	{
		angle = (float) i * d_theta;

		pRing[i].pos.x = ring_radius * ((float)cos( angle ));
		pRing[i].pos.z = ring_radius * ((float)sin( angle ));
		pRing[i].pos.y = 0.0f;

		pRing[i].pos.x += cent_radius;

		// Set normals
		pRing[i].nrm.x = (float) cos( angle );
		pRing[i].nrm.z = (float) sin( angle );
		pRing[i].nrm.y = 0.0f;

		pRing[i].t0.x = (float) i / (float)(ntheta_verts - 1);
	}

	int vind;
	float nx, ny, nz, length;
	vind = 0;

	for( j=0; j < nphi_verts; j++ )
	{
		// rotate each ring into position
		//  around the sweep angle
		
		angle = (float) j * d_phi;

		for( i=0; i < ntheta_verts; i++, vind++ )
		{
			assert( vind < nvert );
			m_pVertices[ vind ].pos.x = pRing[i].pos.x * ((float)cos( angle ));
			m_pVertices[ vind ].pos.y = pRing[i].pos.x * ((float)sin( angle ));
			m_pVertices[ vind ].pos.z = pRing[i].pos.z;

			// rotate normals into position
			nx = m_pVertices[ vind ].nrm.x = pRing[i].nrm.x * ((float)cos(angle));
			ny = m_pVertices[ vind ].nrm.y = pRing[i].nrm.x * ((float)sin(angle));
			nz = m_pVertices[ vind ].nrm.z = pRing[i].nrm.z;

			// normalize normals.  Shouldn't have to in this case, but
			//  just a safety measure.
			length = (float) sqrt(nx*nx + ny*ny + nz*nz);
			assert( length > 0.0f );

			m_pVertices[ vind ].nrm.x = nx / length;
			m_pVertices[ vind ].nrm.y = ny / length;
			m_pVertices[ vind ].nrm.z = nz / length;

			// Set texture coords:
			m_pVertices[ vind ].t0.x = pRing[i].t0.x;
			m_pVertices[ vind ].t0.y = (float) j / (float)( nphi_verts - 1);
		}
	}

	///////////////////////////////////////////
	// Now make indices to form triangles

	int ind;			// index location to fill with a vertex index
	int v1, v2, v3;		// vertex indices of a triangle
	UINT jplus, iplus;	// j+1, i+1 wrapped back to 0 if > Rings or Sides, respectively


	ind = 0;

	for( j=0; j < Rings; j ++ )
	{
		for( i=0; i < Sides; i++ )
		{
			// first triangle of each "square" patch of surface
			//  on the torus

			iplus = i + 1;
			jplus = j + 1;

			// CCW winding order
			v1 = j		* ntheta_verts		+ i;
			v2 = j		* ntheta_verts		+ iplus;
			v3 = jplus	* ntheta_verts		+ i;


			assert( v1 < nvert && v1 >= 0 );
			assert( v2 < nvert && v2 >= 0 );
			assert( v3 < nvert && v3 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;

			// second triangle of the pair
			v1 = v3;
			v2 = v2;
			v3 = jplus * ntheta_verts	+ iplus;

			assert( v3 < nvert && v1 >= 0 );

			assert( ind < nind );
			m_pIndices[ ind++ ] = v1;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v2;
			assert( ind < nind );
			m_pIndices[ ind++ ] = v3;
		}
	}



	//////////////////////////////////////////

	if( pRing != NULL )
	{
		free( pRing );
		pRing = NULL;
	}

	m_bIsValid = true;
	PostInit();

}


void SimpleObject8::InitTorus( float InnerRadius, float OuterRadius, UINT Sides, UINT Rings,
								float TexCoordUMax, float TexCoordVMax )
{

	// Create a torus with user-adjustable scaling of texture coords so a 
	//  texture can repeat several times around the torus

	InitTorus( InnerRadius, OuterRadius, Sides, Rings );

	int i;

	for( i=0; i < m_wNumVerts; i ++ )
	{
		m_pVertices[i].t0.x *= TexCoordUMax;
		m_pVertices[i].t0.y *= TexCoordVMax;
	}

	m_bIsValid = true;
	PostInit();

}


void SimpleObject8::InitClone( SimpleObject8 * pObj )
{
	// copy of the input object

	assert( pObj != NULL );
	assert( pObj->m_bIsValid );

	Allocate( pObj->GetNumVerts(), pObj->GetNumIndices() );

	memcpy( m_pVertices, pObj->m_pVertices, sizeof( SimpleObjVertex ) * m_wNumVerts );
	memcpy( m_pIndices,  pObj->m_pIndices,  sizeof( WORD ) * m_wNumInd );

	m_PrimType = pObj->m_PrimType;
	m_bIsValid = pObj->m_bIsValid;

	PostInit();
}


//////////////////////////////////////////////////////////////////////////////////

void SimpleObject8::AddPositionNoise( D3DXVECTOR3 vNoiseDir, float noise_magnitude )
{
	
	int numvert = m_wNumVerts;

	D3DXVECTOR3 noise;

	for( int ind = 0; ind < numvert; ind++ )
	{
		noise = vNoiseDir * noise_magnitude * ( ((float)rand())/(float)RAND_MAX );
		
		m_pVertices[ ind ].pos.x += noise.x;
		m_pVertices[ ind ].pos.y += noise.y;
		m_pVertices[ ind ].pos.z += noise.z;		
	}
}


void SimpleObject8::AddPositionNoiseGrid( D3DXVECTOR3 vNoiseDir, float noise_magnitude,
										  int grid_nx, int nsmooth )
{
	// Add noise to the positions with possibility for averaging the noise
	//   values with their nearest neighbors as though they lie on a grid 
	//   of x dimension grid_nx.
	// Function does a crude compensation if the number of vertices doesn't fill
	//   out the last line of the imposed grid.
	// noise_magnitude is max magnitude that will be added - the noise field
	//   is scaled so the max value is that.

	int numvert = m_wNumVerts;

	int gridx = grid_nx;
	int gridy = m_wNumVerts/gridx;

	if( gridy * gridx != m_wNumVerts )
	{
		FDebug("SimpleObject8 -- AddPositionNoiseGrid Vertices not a rectangular grid!\n");
		FDebug("    Compensating for remaining verts...\n");

		gridy++;
	}

	D3DXVECTOR3 noise;

	float * noisemag = (float*) calloc( numvert, sizeof( float ));
	assert( noisemag != NULL );

	int ind;

	for( ind = 0; ind < numvert; ind++ )
	{
		noisemag[ind] = ( ((float)rand())/(float)RAND_MAX );
	}

	// now smooth the noise according to the grid
	int i,j, n;
	int ind0, ind1;

	float avg;
	float maxmag;
	float minmag;

	for( n=0; n < nsmooth; n++ )
	{
		maxmag = 0.0f;
		minmag = FLT_MAX;

		for( j=0; j < gridy; j++ )
			for( i=0; i < gridx; i++ )
			{	
				ind0 = j*gridx + i;
				
				if( ind0 > numvert-1 )
				{
					FDebug("broke on:  %d    %d\n", i,j );
					break;
				}

				avg = noisemag[ind0];


				ind1 = ind0 + 1;

				if( ind1 > numvert-1 )
					ind1 = numvert-1;
				avg += noisemag[ind1];


				ind1 = ind0 + gridx;

				if( ind1 > numvert-1 )
				{
					ind1 = i;
				}
				avg += noisemag[ind1];


				ind1 = ind0 + gridx + 1;

				if( ind1 > numvert-1 )
				{
					ind1 = i + 1;
				}
				avg += noisemag[ind1];

				noisemag[ ind0 ] = avg / 4.0f;

				if( noisemag[ ind0 ] > maxmag )
				{
					maxmag = noisemag[ ind0 ];
				}
				if( noisemag[ ind0 ] < minmag )
				{
					minmag = noisemag[ ind0 ];
				}
			}
	}

	// scale the noise values so their max is noise_magnitude
	float scale;
	maxmag = maxmag - minmag;

	if( maxmag > 0.0f )
		scale = noise_magnitude / maxmag;
	else
		assert( false );		// something's gone wrong in the calcs

	for( i=0; i < numvert; i++ )
	{
		noisemag[i] = ( noisemag[i] - minmag ) * scale;
	}


	for( ind = 0; ind < numvert; ind++ )
	{
		noise = vNoiseDir * noisemag[ind];
		
		m_pVertices[ ind ].pos.x += noise.x;
		m_pVertices[ ind ].pos.y += noise.y;
		m_pVertices[ ind ].pos.z += noise.z;		
	}


	free( noisemag );
}


void SimpleObject8::MultiplyTexCoords( float ufactor, float vfactor )
{

	WORD i;

	for( i=0; i < m_wNumVerts; i++ )
	{
		m_pVertices[i].t0.x *= ufactor;
		m_pVertices[i].t0.y *= vfactor;
	}
}


void SimpleObject8::ExplodeFaces( float distance )
{
	// Move faces out along vector from
	//  center of object
	// Does not re-compute face normals

	D3DXVECTOR3 center;

	ComputeObjectCenter( & center );

	int i;
	D3DXVECTOR3 face_center;
	int i1, i2, i3;
	D3DXVECTOR3 vec_from_center;

	// extrude each vertex of each face outward along vector from 
	//  center of object to center of face

	D3DXVECTOR3  * pDispl = new D3DXVECTOR3[ m_wNumVerts ];
	assert( pDispl != NULL );
	memset( pDispl, 0, sizeof( D3DXVECTOR3 ) * m_wNumVerts );

	for( i=0; i < m_wNumInd; i += 3 )
	{
		i1 = m_pIndices[i];
		i2 = m_pIndices[i+1];
		i3 = m_pIndices[i+2];
		
		face_center = ( m_pVertices[i1].pos + m_pVertices[i2].pos + m_pVertices[i3].pos ) / 3.0f;

		vec_from_center = face_center - center;
		D3DXVec3Normalize( &vec_from_center, &vec_from_center );

		pDispl[ i1 ] += vec_from_center;
		pDispl[ i2 ] += vec_from_center;
		pDispl[ i3 ] += vec_from_center;
	}

	for( i=0; i < m_wNumVerts; i ++ )
	{
		D3DXVec3Normalize( & pDispl[i], & pDispl[i] );
		pDispl[i] *= distance;

		m_pVertices[i].pos += pDispl[i];
	}

	SAFE_ARRAY_DELETE( pDispl );
}


void SimpleObject8::ComputeTriCenter( D3DXVECTOR3 * pResult, WORD tri_index )
{
	assert( pResult != NULL );
	tri_index *= 3;
	assert( tri_index + 2  < m_wNumInd );

	*pResult  = m_pVertices[ m_pIndices[ tri_index + 0 ]].pos;
	*pResult += m_pVertices[ m_pIndices[ tri_index + 1 ]].pos;
	*pResult += m_pVertices[ m_pIndices[ tri_index + 2 ]].pos;

	*pResult /= 3.0f;
}


void SimpleObject8::ComputeObjectCenter( D3DXVECTOR3 * pCenter )
{
	assert( pCenter != NULL );

	int i;
	pCenter->x = pCenter->y = pCenter->z = 0.0f;

	for( i=0; i < m_wNumVerts; i++ )
	{
		*pCenter += m_pVertices[i].pos;
	}

	*pCenter = *pCenter / ((float)m_wNumVerts );

}


void SimpleObject8::ReverseNormals()
{
	// invert all normals

	int i;

	for( i=0; i < m_wNumVerts; i++ )
	{
		m_pVertices[i].nrm = - m_pVertices[i].nrm;
	}
}



void SimpleObject8::ReverseWinding()
{
	// change winding for all triangles

	// only knows about reversing trilists for now
	assert( m_PrimType == D3DPT_TRIANGLELIST );


	int i;
	int index;

	// Swap first and second indices for each triangle
	//  to change their windings.

	for( i=0; i < m_wNumInd; i += 3 )
	{
		// += 3 for skipping to next triangle

		index = m_pIndices[ i+1 ];
		m_pIndices[ i+1 ] = m_pIndices[i];
		m_pIndices[i]     = index;
	}
}


void SimpleObject8::SetVertexColors( DWORD ARGB_color )
{
	int ind;

	for( ind = 0; ind < m_wNumVerts; ind ++ )
	{
		m_pVertices[ind].diffuse = ARGB_color;
	}
}


void SimpleObject8::Translate( float x, float y, float z )
{

	int numvert = m_wNumVerts;
	int ind;

	for( ind = 0; ind < numvert; ind ++ )
	{
		m_pVertices[ind].pos.x += x;		
		m_pVertices[ind].pos.y += y;		
		m_pVertices[ind].pos.z += z;		
	}
}


void SimpleObject8::TranslateXMinToZero()
{
	int numvert = m_wNumVerts;

	float minx = FLT_MAX;
	int ind;

	for( ind = 0; ind < numvert; ind++ )
	{
		if( m_pVertices[ ind ].pos.x < minx )
			minx = m_pVertices[ ind ].pos.x;
	}

	for( ind = 0; ind < numvert; ind++ )
	{
		m_pVertices[ ind ].pos.x -= minx;
	}
}


void SimpleObject8::GenerateNormalsCCW()
{
	// Generate normals for counter-clockwise right-handed coordinate
	//   system models

	GenerateNormals();

	// Reverse their direction:
	int i;
	for( i=0; i < m_wNumVerts; i++ )
	{
		m_pVertices[i].nrm = - m_pVertices[i].nrm;
	}
}



void SimpleObject8::GenerateNormals()
{
	assert( m_pIndices != NULL );
	assert( m_pVertices != NULL );

	if( m_PrimType != D3DPT_TRIANGLELIST )
	{
		FDebug("Error: primitive type must be D3DPT_TRIANGLELIST\n");
		assert( false );
	}


	int i;
	int vi0, vi1, vi2;	    // vertex indices of triangle

	// zero normals for accumulation
	for( i=0; i < m_wNumVerts; i++ )
	{
		m_pVertices[i].nrm.x = 0.0f;
		m_pVertices[i].nrm.y = 0.0f;
		m_pVertices[i].nrm.z = 0.0f;
	}

	// find each triangle's normal and add into the 
	//   vertex normal

	D3DXVECTOR3 v0, v1, v2;
	D3DXVECTOR3 d0, d1, nrm;

	i=0;
	while( i < m_wNumInd )
	{
		vi0 = m_pIndices[i++];		
		vi1 = m_pIndices[i++];		
		vi2 = m_pIndices[i++];		

		v0 = m_pVertices[ vi0 ].pos;
		v1 = m_pVertices[ vi1 ].pos;
		v2 = m_pVertices[ vi2 ].pos;

		d0 = v1 - v0;
		d1 = v2 - v0;

		D3DXVec3Cross( &nrm, &d1, &d0 );		// triangle normal

		//  accumulate into vertex normals:

		m_pVertices[ vi0 ].nrm += nrm;
		m_pVertices[ vi1 ].nrm += nrm;
		m_pVertices[ vi2 ].nrm += nrm;
	}

	// normalize the normals
	for( i=0; i < m_wNumVerts; i++ )
	{
		D3DXVec3Normalize( &( m_pVertices[i].nrm ), &( m_pVertices[i].nrm ) );
	}
}







void SimpleObject8::Draw( LPDIRECT3DDEVICE8  pD3DDev )
{
	assert( pD3DDev != NULL );

	FDebug("SimpleObject8::  Non-VertexBuffer rendering!  This will be slow!\n");

	pD3DDev->DrawIndexedPrimitiveUP(	m_PrimType,
										0,				// min index
										m_wNumVerts,
										m_wNumInd/3,	// prim count
										m_pIndices,
										D3DFMT_INDEX16,
										m_pVertices,
										sizeof( SimpleObjVertex ) );


}

